import json
import os

def generate_index_report(data, report_name="Index Report",filename='index_report'):
    """
    Generate a JSON report of index information.

    Parameters:
        data (list of lists): Raw index data where each inner list contains index properties.
        report_name (str): Name of the report.

    Returns:
        str: JSON formatted string representing the index report.
    """
    headers = ['Database Name', 'Index Name', 'Category']
    indexes = [dict(zip(headers, row)) for row in data]
    report = {
        "report_name": report_name,
        "database_name": os.getenv('DB_NAME'),
        "total_index_count": len(indexes),
        "indexes": indexes
    }
    with open(f'''{filename}.json''','w') as output_json:
        json.dump(report,output_json,indent=4)
        return True

def generate_command(category,schema_name,index_name):
    
    if category=="Bloated":
        execute_sql=f'''REINDEX INDEX CONCURRENTLY {schema_name}.{index_name};'''
        return execute_sql
    else:
        execute_sql=f'''DROP INDEX CONCURRENTLY {schema_name}.{index_name};'''
        return execute_sql


class SqlQueries:
    @staticmethod
    def find_unused_redundant_indexes():
        """Returns a query to find unused and redundant indexes."""
        return """
        WITH unused_indexes AS (
            SELECT
                i.relname AS index_name,
                t.relname AS table_name,
                pg_size_pretty(pg_relation_size(i.oid)) AS index_size,
                s.idx_scan AS index_scans
            FROM
                pg_stat_user_indexes AS s
            JOIN
                pg_index AS idx ON s.indexrelid = idx.indexrelid
            JOIN
                pg_class AS i ON i.oid = s.indexrelid
            JOIN
                pg_class AS t ON t.oid = idx.indrelid
            WHERE
                s.idx_scan = 0
                AND (last_idx_scan IS NULL OR last_idx_scan < NOW() - INTERVAL '1 year')
                AND NOT idx.indisprimary
                AND NOT idx.indisunique
        ),
        redundant_indexes AS (
             SELECT
            tnsp.nspname AS schema_name,
            trel.relname AS table_name,
            irel.relname AS index_name,
            string_agg(a.attname, ', ' ORDER BY c.ordinality) AS columns
          FROM pg_index AS i
          JOIN pg_class AS trel ON trel.oid = i.indrelid
          JOIN pg_namespace AS tnsp ON trel.relnamespace = tnsp.oid
          JOIN pg_class AS irel ON irel.oid = i.indexrelid
          JOIN pg_attribute AS a ON trel.oid = a.attrelid
          JOIN LATERAL unnest(i.indkey) 
            WITH ORDINALITY AS c(colnum, ordinality)
              ON a.attnum = c.colnum
          GROUP BY i, tnsp.nspname, trel.relname, irel.relname
        ),
        redundant_list AS(
        SELECT
          i.schema_name AS "schema_name",
          i.table_name,
          i.index_name AS "index_name",
          i.columns AS "columns_list",
          j.index_name AS "existsing_index",
          j.columns AS "existing_columns"
        FROM redundant_indexes i
        JOIN redundant_indexes j
          ON i.schema_name = j.schema_name
          AND i.table_name = j.table_name
          AND j.columns LIKE i.columns || ',%'
        )
        SELECT
            r.schema_name AS schema_name,
            u.table_name AS table_name,
            u.index_name AS index_name,
            u.index_scans AS index_scans,
            u.index_size AS index_size
        FROM
            unused_indexes u
        JOIN
            redundant_list r ON u.index_name = r.index_name;
        """

    @staticmethod
    def find_unused_indexes():
        """Returns a query to list all indexes."""
        return """
            SELECT
                s.schemaname AS schema_name,
                t.relname AS table_name,
                i.relname AS index_name,
                idx_scan AS index_scans,
                pg_size_pretty(pg_relation_size(i.oid)) AS index_size,
                last_idx_scan AS last_scan_date
            FROM
                pg_stat_user_indexes AS s
            JOIN
                pg_index AS idx ON s.indexrelid = idx.indexrelid
            JOIN
                pg_class AS i ON i.oid = s.indexrelid
            JOIN
                pg_class AS t ON t.oid = idx.indrelid
            WHERE
                s.idx_scan=0
                AND (last_idx_scan IS NULL OR last_idx_scan < NOW() - INTERVAL '1 year')
                AND NOT idx.indisprimary
                AND NOT idx.indisunique
            ORDER BY
                idx_scan DESC;
        """

    @staticmethod
    def find_last_usage_older_than_one_year_indexes():
        """Returns a query to list all indexes which scanned over last year"""
        return """
            SELECT
                s.schemaname AS schema_name,
                t.relname AS table_name,
                i.relname AS index_name,
                idx_scan AS index_scans,
                pg_size_pretty(pg_relation_size(i.oid)) AS index_size
        FROM
            pg_stat_user_indexes AS s
        JOIN
            pg_index AS idx ON s.indexrelid = idx.indexrelid
        JOIN
            pg_class AS i ON i.oid = s.indexrelid
        JOIN
            pg_class AS t ON t.oid = idx.indrelid
        WHERE
            (last_idx_scan IS NULL OR last_idx_scan < NOW() - INTERVAL '1 year')
            AND NOT idx.indisprimary
            AND NOT idx.indisunique
        ORDER BY
            idx_scan DESC;
    """

    @staticmethod
    def find_invalid_indexes():
        """Returns a query to list all indexes which scanned over last year"""
        return """
            SELECT
                s.schemaname AS schema_name,
                t.relname AS table_name,
                i.relname AS index_name,
                idx_scan AS index_scans,
                pg_size_pretty(pg_relation_size(i.oid)) AS index_size
            FROM
                pg_stat_user_indexes AS s
            JOIN
                pg_index AS idx ON s.indexrelid = idx.indexrelid
            JOIN
                pg_class AS i ON i.oid = s.indexrelid
            JOIN
                pg_class AS t ON t.oid = idx.indrelid WHERE idx.indisvalid is FALSE;
    """

    @staticmethod
    def calculate_btree_bloat():
        return """
    SELECT current_database(), 
       nspname AS schemaname, 
       tblname, 
       idxname, 
       bs * (relpages)::bigint AS real_size,
       bs * (relpages - est_pages)::bigint AS extra_size,
       100 * (relpages - est_pages)::float / relpages AS extra_pct,
       fillfactor,
       CASE WHEN relpages > est_pages_ff 
            THEN bs * (relpages - est_pages_ff) 
            ELSE 0 
       END AS bloat_size,
       100 * (relpages - est_pages_ff)::float / relpages AS bloat_pct,
       is_na
FROM (
    SELECT coalesce(1 + ceil(reltuples / floor((bs - pageopqdata - pagehdr) / (4 + nulldatahdrwidth)::float)), 0) AS est_pages,
           coalesce(1 + ceil(reltuples / floor((bs - pageopqdata - pagehdr) * fillfactor / (100 * (4 + nulldatahdrwidth)::float))), 0) AS est_pages_ff,
           bs, 
           nspname, 
           tblname, 
           idxname, 
           relpages, 
           fillfactor, 
           is_na
    FROM (
        SELECT maxalign, 
               bs, 
               nspname, 
               tblname, 
               idxname, 
               reltuples, 
               relpages, 
               idxoid, 
               fillfactor,
               (index_tuple_hdr_bm + maxalign - CASE WHEN index_tuple_hdr_bm % maxalign = 0 THEN maxalign ELSE index_tuple_hdr_bm % maxalign END + nulldatawidth + maxalign - CASE WHEN nulldatawidth = 0 THEN 0 WHEN nulldatawidth::integer % maxalign = 0 THEN maxalign ELSE nulldatawidth::integer % maxalign END) AS nulldatahdrwidth, 
               pagehdr, 
               pageopqdata, 
               is_na
        FROM (
            SELECT n.nspname, 
                   i.tblname, 
                   i.idxname, 
                   i.reltuples, 
                   i.relpages, 
                   i.idxoid, 
                   i.fillfactor,
                   current_setting('block_size')::numeric AS bs,
                   CASE WHEN version() ~ 'mingw32' OR version() ~ '64-bit|x86_64|ppc64|ia64|amd64' THEN 8 ELSE 4 END AS maxalign,
                   24 AS pagehdr,
                   16 AS pageopqdata,
                   CASE WHEN max(coalesce(s.null_frac, 0)) = 0 THEN 8 ELSE 8 + ((32 + 8 - 1) / 8) END AS index_tuple_hdr_bm,
                   sum((1 - coalesce(s.null_frac, 0)) * coalesce(s.avg_width, 1024)) AS nulldatawidth,
                   max(CASE WHEN i.atttypid = 'pg_catalog.name'::regtype THEN 1 ELSE 0 END) > 0 AS is_na
            FROM (
                SELECT ct.relname AS tblname, 
                       ct.relnamespace, 
                       ic.idxname, 
                       ic.attpos, 
                       ic.indkey, 
                       ic.indkey[ic.attpos], 
                       ic.reltuples, 
                       ic.relpages, 
                       ic.tbloid, 
                       ic.idxoid, 
                       ic.fillfactor,
                       coalesce(a1.attnum, a2.attnum) AS attnum, 
                       coalesce(a1.attname, a2.attname) AS attname, 
                       coalesce(a1.atttypid, a2.atttypid) AS atttypid,
                       CASE WHEN a1.attnum IS NULL THEN ic.idxname ELSE ct.relname END AS attrelname
                FROM (
                    SELECT idxname, 
                           reltuples, 
                           relpages, 
                           tbloid, 
                           idxoid, 
                           fillfactor, 
                           indkey, 
                           pg_catalog.generate_series(1, indnatts) AS attpos
                    FROM (
                        SELECT ci.relname AS idxname, 
                               ci.reltuples, 
                               ci.relpages, 
                               i.indrelid AS tbloid, 
                               i.indexrelid AS idxoid,
                               coalesce(substring(array_to_string(ci.reloptions, ' ') from 'fillfactor=([0-9]+)')::smallint, 90) AS fillfactor,
                               i.indnatts, 
                               pg_catalog.string_to_array(pg_catalog.textin(pg_catalog.int2vectorout(i.indkey)), ' ')::int[] AS indkey
                        FROM pg_catalog.pg_index i
                        JOIN pg_catalog.pg_class ci ON ci.oid = i.indexrelid
                        WHERE ci.relam = (SELECT oid FROM pg_am WHERE amname = 'btree') 
                          AND ci.relpages > 0
                    ) AS idx_data
                ) AS ic
                JOIN pg_catalog.pg_class ct ON ct.oid = ic.tbloid
                LEFT JOIN pg_catalog.pg_attribute a1 ON ic.indkey[ic.attpos] <> 0 AND a1.attrelid = ic.tbloid AND a1.attnum = ic.indkey[ic.attpos]
                LEFT JOIN pg_catalog.pg_attribute a2 ON ic.indkey[ic.attpos] = 0 AND a2.attrelid = ic.idxoid AND a2.attnum = ic.attpos
            ) i
            JOIN pg_catalog.pg_namespace n ON n.oid = i.relnamespace
            JOIN pg_catalog.pg_stats s ON s.schemaname = n.nspname AND s.tablename = i.attrelname AND s.attname = i.attname
            WHERE n.nspname NOT IN ('pg_catalog', 'information_schema')
            GROUP BY 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11
        ) AS rows_data_stats
    ) AS rows_hdr_pdg_stats
) AS relation_stats
ORDER BY nspname, tblname, idxname;
    """

    @staticmethod
    def find_exact_duplicate_index():
        """Returns a query to list all indexes which scanned over last year"""
        return """
            SELECT 
                   (array_agg(idx))[1] as index_name_1, 
                   (array_agg(idx))[2] as index_name_2,
                   pg_size_pretty(sum(pg_relation_size(idx))::bigint) as size
            FROM (
                SELECT indexrelid::regclass as idx, (indrelid::text ||E'\n'|| indclass::text ||E'\n'|| indkey::text ||E'\n'||
                                                     coalesce(indexprs::text,'')||E'\n' || coalesce(indpred::text,'')) as key
                FROM pg_index) sub
            GROUP BY key HAVING count(*)>1
            ORDER BY sum(pg_relation_size(idx)) DESC;
    """

    @staticmethod
    def find_duplicate_constraints():
        """Returns a list of unique indexes which are valid,ready and unique"""
        return """
            SELECT
                ix.schemaname,
                ix.tablename,
                ix.indexname,
                ix.indexdef,
                pg_size_pretty(pg_relation_size(cs.oid)) AS index_size
            FROM
                pg_stat_user_indexes co
                INNER JOIN pg_indexes ix ON co.indexrelname = ix.indexname
                INNER JOIN pg_index i ON co.indexrelid = i.indexrelid
                INNER JOIN pg_class AS cs ON cs.oid = co.indexrelid
            WHERE
                i.indisunique = true
                AND i.indisvalid = true
                AND i.indisready = true
            ORDER BY
                ix.tablename;
    """
    
    @staticmethod
    def find_duplicate_btrees():
        """Returns a list of unique indexes which are valid,ready and unique"""
        return """
            SELECT
                ix.schemaname,
                ix.tablename,
                ix.indexname,
                ix.indexdef,
                pg_size_pretty(pg_relation_size(cs.oid)) AS index_size
            FROM
                pg_stat_user_indexes co
                INNER JOIN pg_indexes ix ON co.indexrelname = ix.indexname
                INNER JOIN pg_index i ON co.indexrelid = i.indexrelid
                INNER JOIN pg_class AS cs ON cs.oid = co.indexrelid
            WHERE
                i.indisunique = false
                AND i.indisvalid = true
                AND i.indisready = true
            ORDER BY
                ix.tablename;
    """




import os
import psycopg2
from .queries import SqlQueries
import logging

class DatabaseManager:
    """
    A class to manage database connections and operations related to PostgreSQL.

    This class handles connecting to the PostgreSQL database, retrieving information 
    about indexes (such as unused, invalid, duplicate, and bloated indexes), and 
    collecting facts about the database's state (like recovery status and replication).

    Attributes:
        connection (psycopg2.connection): The connection object for the PostgreSQL database.
        replica_node_exists (bool): Indicates if a replica node exists.
        recovery_status (bool): The recovery status of the database.

    Methods:
        connect(): Establishes a database connection using environment variables.
        close(): Closes the database connection.
        collect_facts(): Collects and stores facts about the database's state.
        get_unused_and_invalid_indexes(): Retrieves unused, invalid, and duplicate indexes.
        get_bloated_indexes(): Identifies bloated B-tree indexes in the database.
        fetch_invalid_indexes(): Identifies invalid indexes that require attention.
        fetch_unused_indexes(): Retrieves indexes that have not been used in a specified timeframe.
    """
    logger = logging.getLogger("pgindexinsight")
    logger.setLevel(logging.WARNING)  # Set default logging level to DEBUG
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING)  # Set console log level (can adjust to DEBUG if needed)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    MIN_SUPPORTED_VERSION = 16
    SYSTEM_DATABASE_LIST = ['postgres','template0','template1']
    def __init__(self):
        self.connection = None
        self.replica_node_exists = None
        self.recovery_status = None
        self.database_version=None
        self.collect_facts()

    def connect(self):
        """Initializes the DatabaseManager and collects database facts."""
        if not (os.getenv('DB_NAME') not in DatabaseManager.SYSTEM_DATABASE_LIST):
            raise ValueError(f"System databases is not allowed to be analyzed")
        if self.connection is None:
            try:
                host = os.getenv("DB_HOST", "localhost")
                port = os.getenv("DB_PORT", "5432")
                dbname = os.getenv("DB_NAME")
                user = os.getenv("DB_USER")
                password = os.getenv("DB_PASSWORD")

                if not all([dbname, user, password]):
                    raise ValueError(
                        "Missing one or more required environment variables: DB_NAME, DB_USER, DB_PASSWORD"
                    )
                self.connection = psycopg2.connect(
                    host=host,
                    port=port,
                    dbname=dbname,
                    user=user,
                    password=password,
                    connect_timeout=10,
                    options="-c statement_timeout=5000",
                    application_name="pgindexinsight",
                )
                self.connection.autocommit = True
                self.check_superuser()
            except Exception as e:
                raise ConnectionError(f"Error connecting to the database: {str(e)}")

        return self.connection

    def check_superuser(self):
        """Checks if the connected user is a superuser and logs a debug message."""
        try:
            with self.connection.cursor() as cur:
                cur.execute("SELECT current_setting('is_superuser')")
                is_superuser = cur.fetchone()[0]
                if is_superuser == 'on':
                    DatabaseManager.logger.warning("Connected as a superuser.")
                else:
                    DatabaseManager.logger.info("Connected as a regular user.")
        except Exception as e:
            DatabaseManager.logger.error(f"Failed to check superuser status: {e}")


    def close(self):
        """Closes the database connection."""
        if self.connection:
            self.connection.close()
            self.connection = None

    def collect_facts(self):
        """Collects and sets database recovery and replication status."""
        database_connection = self.connect()
        with database_connection.cursor() as db_cursor:
            db_cursor.execute("select pg_is_in_recovery()")
            recovery_status = db_cursor.fetchall()
            recovery_status = recovery_status[0][0]
            self.recovery_status = recovery_status
            db_cursor.execute(
                f"""select count(*) as physical_repl_count from pg_replication_slots where slot_type='physical' and active is true """
            )
            replica_count = db_cursor.fetchall()
            replica_count = replica_count[0][0]
            if replica_count > 0:
                self.replica_node_exists = True
            else:
                self.replica_node_exists = False
            db_cursor.execute('select version()')
            database_version=db_cursor.fetchall()
            database_version=float(str(database_version[0][0]).split(' ')[1])
            self.database_version=database_version
    
    def _check_version_supported(self):
        """Ensures that the database version is supported."""
        if self.database_version < self.MIN_SUPPORTED_VERSION:
            raise ValueError(f"PostgreSQL version {self.MIN_SUPPORTED_VERSION}.0 and higher is supported.")

    
    def get_unused_and_invalid_indexes(self):
        """Retrieves a list of unused, invalid, and duplicate indexes in the database."""
        self._check_version_supported()
        try:
            conn = self.connect()

            with conn.cursor() as cur:
                final_result = []

                cur.execute(SqlQueries.find_unused_redundant_indexes())
                unused_redundant_result = cur.fetchall()
                for row in unused_redundant_result:
                    final_result.append(
                        {
                            "database_name": os.getenv("DB_NAME"),
                            "schema_name": row[0],
                            "index_name": row[2],
                            "index_size": row[4],
                            "category": "Unused&Redundant Index",
                        }
                    )

                cur.execute(SqlQueries.find_invalid_indexes())
                invalid_result = cur.fetchall()
                for row in invalid_result:
                    final_result.append(
                        {
                            "database_name": os.getenv("DB_NAME"),
                            "schema_name": row[0],
                            "index_name": row[2],
                            "index_size": row[4],
                            "category": "Invalid Index",
                        }
                    )
                if len(final_result) == 0:
                    return []
                return final_result

        except Exception as e:
            print(f"No Result, Failed due to: {e}")
        finally:
            self.close()

    def get_bloated_indexes(self,bloat_threshold):
        """Finds duplicate B-tree indexes in the database."""
        self._check_version_supported()
        try:
            conn = self.connect()
            with conn.cursor() as cur:
                cur.execute(SqlQueries.calculate_btree_bloat())
                bloated_indexes = cur.fetchall()
                bloatedIndexList = []
                for index in bloated_indexes:
                    indexModel = {
                        "database_name": index[0],
                        "schema_name": index[1],
                        "index_name": index[3],
                        "bloat_ratio": float(format(index[9], ".1f")),
                        "category": "Bloated",
                    }
                    if indexModel.get("bloat_ratio") > bloat_threshold:
                        bloatedIndexList.append(indexModel)
                return bloatedIndexList

        except Exception as e:
            print(f"No Result, Failed due to: {e}")
        finally:
            self.close()

    def fetch_invalid_indexes(self):
        """Identifies invalid indexes that may need to be cleaned or rebuilt."""
        self._check_version_supported()
        database_connection = self.connect()
        with database_connection.cursor() as database_cursor:
            database_cursor.execute(SqlQueries.find_invalid_indexes())
            invalid_indexes = database_cursor.fetchall()
            invalid_index_list = []
            for index in invalid_indexes:
                invalid_index_dict = {
                    "database_name": os.getenv("DB_NAME"),
                    "schema_name": index[0],
                    "index_name": index[2],
                    "index_size": index[4],
                    "category": "Invalid Index.",
                }
                invalid_index_list.append(invalid_index_dict)

        return invalid_index_list

    def fetch_unused_indexes(self):
        """Retrieves indexes that have not been used in over a specified timeframe."""
        self._check_version_supported()
        database_connection = self.connect()
        with database_connection.cursor() as database_cursor:
            database_cursor.execute(SqlQueries.find_unused_indexes())
            old_indexes = database_cursor.fetchall()
            old_index_list = []
            for index in old_indexes:
                old_index_dict = {
                    "database_name": os.getenv("DB_NAME"),
                    "schema_name": index[0],
                    "index_name": index[2],
                    "index_size": index[4],
                    "index_scan": index[3],
                    "last_scan": index[5],
                    "category": "Unused/Retired Index",
                }
                old_index_list.append(old_index_dict)
        return old_index_list

    def fetch_duplicate_unique_indexes(self):
        """Retrieves unique indexes have being duplicated"""
        self._check_version_supported()
        database_connection = self.connect()
        current_indexes = set()
        duplicate_unique_indexes = []
        with database_connection.cursor() as database_cursor:
            database_cursor.execute(SqlQueries.find_duplicate_constraints())
            unique_indexes = database_cursor.fetchall()
            for index in unique_indexes:
                index_columns=str(index[3]).split(' ')[8]
                schema_name=index[0]
                table_name=index[1]
                index_record=(schema_name,table_name,index_columns)
                if index_record in current_indexes:
                    # if index record has been found in current_indexes list append index to duplicate_unique_indexes list.
                    duplicate_unique_indexes.append(index)
                else:
                    # if index record has not been found in current indexes add index_record to current_indexes list to 
                    # compare later.
                    current_indexes.add(index_record)
        return duplicate_unique_indexes

    def fetch_duplicate_indexes(self):
        """Retrieves unique indexes have being duplicated"""
        self._check_version_supported()
        database_connection = self.connect()
        current_indexes = set()
        duplicate_unique_indexes = []
        with database_connection.cursor() as database_cursor:
            database_cursor.execute(SqlQueries.find_duplicate_btrees())
            unique_indexes = database_cursor.fetchall()
            for index in unique_indexes:
                index_columns=str(index[3]).split(' ')[7]
                schema_name=index[0]
                table_name=index[1]
                index_record=(schema_name,table_name,index_columns)
                if index_record in current_indexes:
                    # if index record has been found in current_indexes list append index to duplicate_unique_indexes list.
                    duplicate_unique_indexes.append(index)
                else:
                    # if index record has not been found in current indexes add index_record to current_indexes list to 
                    # compare later.
                    current_indexes.add(index_record)
        return duplicate_unique_indexes


from setuptools import setup, find_packages

setup(
    name='pg_index_insight',
    version='v1.3.0',
    author='Huseyin Demir',
    author_email='huseyin.d3r@gmail.com',
    description='A CLI tool for analyzing PostgreSQL index efficiency',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/kylorend3r/pg_index_insight',  
    packages=find_packages(), 
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'click', 
        'psycopg2==2.9.9',
        'tabulate'
    ],
    entry_points={
        'console_scripts': [
            'pgindexinsight=pg_index_insight.cli:main',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import click
import os
from tabulate import tabulate
import time
from .utils import generate_index_report
from .utils import generate_command
from .database import DatabaseManager as DatabaseManager

@click.command()
@click.option("--json", is_flag=True, help="Export output to JSON file.")
def list_unused_or_old_indexes(json):
    """
    Connects to the PostgreSQL database and retrieves unused or redundant indexes. 
    This function queries the database for indexes that are not frequently scanned 
    or are outdated, indicating that they may be redundant. If such indexes are found, 
    it displays them in a tabular format with details including the database name, 
    schema name, index name, size, scan count, last scan date, and other related 
    attributes such as replica status and recovery mode.

    If no unused or redundant indexes are found, the function informs the user 
    and exits gracefully.
    """
    try:
        database_instance = DatabaseManager()
        duplicate_index_list = database_instance.fetch_unused_indexes()
        database_name=os.getenv('DB_NAME')
        report_time = str.replace(str(time.time()), ".", "_")
        json_report_name=f'''{database_name}_unused_old_index_{report_time}'''
        if not len(duplicate_index_list)>0:
            click.echo(f'No unused or old index found for database: {database_name}')
            exit(0)
        table_formatted_index_result = [
            [
                item["database_name"],
                item["schema_name"],
                item["index_name"],
                item["index_size"],
                item["index_scan"],
                item["last_scan"],
                item["category"],
                database_instance.replica_node_exists,
                database_instance.recovery_status,
            ]
            for item in duplicate_index_list
        ]
        index_table_headers = [
            "Database Name",
            "Schema Name",
            "Index Name",
            "Index Size",
            "Index Scan Count",
            "Last Scan Date",
            "Category",
            "Replica Node Exists",
            "Database Recovery Mode"
        ]
        index_result_table = tabulate(
            table_formatted_index_result, index_table_headers, tablefmt="psql"
        )
        if json:
            jsonReport = generate_index_report(
                table_formatted_index_result, filename=json_report_name
            )
            if not jsonReport:
                click.echo(f"Failed to export json")
        click.echo(index_result_table)
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@click.command()
@click.option('--dry-run', is_flag=True, help="Perform a dry run without making any changes.")
@click.option("--json", is_flag=True, help="Export output to JSON file.")
def list_invalid_indexes(dry_run,json):
    """
    Connects to the PostgreSQL database and retrieves invalid indexes. 
    Invalid indexes typically refer to indexes that are misconfigured, 
    corrupted, or otherwise ineffective in optimizing queries.

    This function queries the database for such invalid indexes and, 
    if found, displays them in a table with details including the database name, 
    schema name, index name, size, and a category indicating the type of invalidity.

    If no invalid indexes are found, the function informs the user and exits.
    """

    try:
        database_query=DatabaseManager()
        invalid_indexes=database_query.fetch_invalid_indexes()
        database_name=os.getenv('DB_NAME')
        report_time = str.replace(str(time.time()), ".", "_")
        json_report_name=f'''{database_name}_invalid_index_{report_time}'''
        if not len(invalid_indexes)>0:
            click.echo(f'No invalid index found for database: {database_name}')
            exit(0)
       
        if not len(invalid_indexes)==0:
            
            table_formatted_index_result = [
                [
                    item["database_name"],
                    item["schema_name"],
                    item["index_name"],
                    item["index_size"],
                    item["category"],
                ]
                for item in invalid_indexes
            ]
            index_table_headers = [
                "Database Name",
                "Schema Name",
                "Index Name",
                "Index Size",
                "Category",
            ]
            index_result_table = tabulate(
                table_formatted_index_result, index_table_headers, tablefmt="psql"
            )
            click.echo(index_result_table)
            if json:
                jsonReport = generate_index_report(
                    table_formatted_index_result, filename=json_report_name
            )
            if not jsonReport:
                click.echo(f"Failed to export json")
            if dry_run:
                click.echo(f'''The following queries might be run on database: {database_name} to remove invalid indexes. Please run the commands wisely.''')
                for index in invalid_indexes:
                    command_executed=generate_command(index['category'],index['schema_name'],index['index_name'])
                    click.echo(command_executed)
        
    except Exception as e:
        click.echo(f"Error: {str(e)}")


@click.command()
@click.option("--json", is_flag=True, help="Export output to JSON file.")
@click.option('--dry-run', is_flag=True, help="Perform a dry run without making any changes.")
def list_unemployed_indexes(json,dry_run):
    """
    Connects to the PostgreSQL database and identifies inefficient indexes, 
    which may include unused or invalid indexes that do not contribute to query 
    performance effectively. 

    This function retrieves such indexes from the database and formats the 
    results into a table displaying the database name, index name, and category. 
    If the user requests a JSON export, the function generates a report with 
    the results, naming the file based on the current time and database name.

    If no inefficient indexes are found, the function informs the user and exits. 
    If the JSON report generation fails, a corresponding error message is displayed.
    
    Parameters:
        json (bool): A flag indicating whether to export the results as a JSON report.
    """
    try:
        database_query = DatabaseManager()
        indexResult = database_query.get_unused_and_invalid_indexes()
        duplicate_unique_indexes_result=database_query.fetch_duplicate_unique_indexes() 
        duplicate_btree_indexes_result=database_query.fetch_duplicate_indexes()
        database_name=os.getenv('DB_NAME')
        if len(indexResult)==0 and len(duplicate_unique_indexes_result)==0 and len(duplicate_btree_indexes_result)==0:
            click.echo(f'No inefficient index found for database: {database_name}')
            exit(0)
        table_formatted_index_result = [
            [item["database_name"], item["schema_name"],item["index_name"],item["index_size"], item["category"]]
            for item in indexResult
        ]
        # append duplicate unique indexes
        for unique_index in duplicate_unique_indexes_result:
            table_formatted_index_result.append([
                database_name,
                unique_index[0],
                unique_index[2],
                unique_index[4],
                "Duplicate Unique Index"
            ])

        # append duplicate btree indexes
        for btree_index in duplicate_btree_indexes_result:
            table_formatted_index_result.append([
                database_name,
                btree_index[0],
                btree_index[2],
                btree_index[4],
                "Duplicate Btree Index"
            ])
        index_table_headers = ["Database Name","Schema Name", "Index Name","Index Size", "Category"]
        report_time = str.replace(str(time.time()), ".", "_")
        json_report_name=f'''{database_name}_inefficient_index_{report_time}'''
        sorted_desc_index_list = sorted(table_formatted_index_result, key=lambda x: x[3],reverse=True)
        index_result_table = tabulate(
            sorted_desc_index_list, index_table_headers, tablefmt="psql"
        )
        if json:
            jsonReport = generate_index_report(
                sorted_desc_index_list, filename=json_report_name
            )
            if not jsonReport:
                click.echo(f"Failed to export json")
        click.echo(index_result_table)
        if dry_run:
            click.echo(f'''The following queries might be run on database: {database_name}. Please run the commands wisely.''')
            for index in indexResult:
                command_executed=generate_command(index['category'],index['schema_name'],index['index_name'])
                click.echo(command_executed)
            for index in duplicate_unique_indexes_result:
                command_executed=generate_command(index[3],index[0],index[2])
                click.echo(command_executed)
            
        
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@click.command()
@click.option("--json", is_flag=True, help="Export output to JSON file.")
@click.option('--dry-run', is_flag=True, help="Perform a dry run without making any changes.")
@click.option('--bloat-threshold', type=int, default=50, help="Set the bloat threshold percentage for indexes.")
def list_bloated_btree_indexes(json,dry_run,bloat_threshold):
    """
    Connects to the PostgreSQL database and identifies bloated B-tree indexes. 
    Bloated indexes occur when the index structure has a significant amount of 
    unused space, which can lead to inefficient query performance and wasted 
    storage resources.

    This function retrieves bloated index information and formats the results 
    into a table displaying the database name, index name, and category. 
    If the user requests a JSON export, the function generates a report with 
    the results, naming the file based on the current time and database name.

    If no bloated indexes are found, the function informs the user and exits. 
    If the JSON report generation fails, a corresponding error message is displayed.

    Parameters:
        json (bool): A flag indicating whether to export the results as a JSON report.
    """
    try:
        databaseConnection = DatabaseManager()
        indexResult = databaseConnection.get_bloated_indexes(bloat_threshold)
        database_name=os.getenv('DB_NAME')
        if not len(indexResult)>0:
            click.echo(f'No bloated index found for database: {database_name}')
            exit(0)
        table_formatted_index_result = [
            [item["database_name"],item["schema_name"], item["index_name"],item["bloat_ratio"],item["category"]]
            for item in indexResult
        ]
        index_table_headers = ["Database Name","Schema Name","Index Name","Bloat Ratio", "Category"]
        report_time = str.replace(str(time.time()), ".", "_")
        json_report_name=f'''{database_name}_bloated_index_{report_time}'''
        index_result_table = tabulate(
            table_formatted_index_result, index_table_headers, tablefmt="psql"
        )
        if json:
            jsonReport = generate_index_report(
                table_formatted_index_result, filename=json_report_name
            )
            if not jsonReport:
                click.echo(f"Failed to export json")
        click.echo(index_result_table)

        if dry_run:
            click.echo(f'''The following queries might be run on database: {database_name}. Please run the commands wisely.''')
            for index in indexResult:
                command_executed=generate_command(index['category'],index['schema_name'],index['index_name'])
                click.echo(command_executed)
        
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@click.group()
def main():
    """
    The main entry point for the pgindexinsight CLI Tool. 

    This command line interface provides various utilities to analyze and 
    manage PostgreSQL indexes, helping users identify and eliminate 
    inefficient, invalid, unused, or bloated indexes. 

    Available commands include:
    - list_unused_or_old_indexes: Lists indexes that are no longer in use.
    - list_invalid_indexes: Identifies indexes that are misconfigured or corrupted.
    - list_duplicate_indexes: Finds duplicate B-tree indexes.
    - list_unemployed_indexes: Reports on indexes that are underperforming.
    - list_bloated_btree_indexes: Detects indexes with excessive unused space.

    To use this tool, invoke it from the command line and specify a command.
    """
    pass


main.add_command(list_bloated_btree_indexes)
main.add_command(list_unemployed_indexes)
main.add_command(list_invalid_indexes)
main.add_command(list_unused_or_old_indexes)

if __name__ == '__main__':
    main()
